#include <stdio.h>
#include <stdlib.h>

#include <string>
#include <vector>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/select.h>

#include "parser.h"
#include "log.h"
#include "rover.h"

using namespace std;

bool come_tel = false;

bool done = false;

NetStream::NetStream(const char* host, const char* port) {
    LOG("Connecting to %s:%s\n", host, port);

    readFd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
    CHECK(readFd >= 0);

    struct hostent* he = gethostbyname(host);

    struct sockaddr_in addr;
    memset(&addr, 0, sizeof(addr));
    addr.sin_family = PF_INET;
    addr.sin_addr.s_addr = *(in_addr_t*)he->h_addr_list[0];
    addr.sin_port = htons(atoi(port));

    CHECK(connect(readFd, (struct sockaddr*)&addr, sizeof(addr)) >= 0);

    int nodelay = 1;
    CHECK(setsockopt(readFd, IPPROTO_TCP, TCP_NODELAY, &nodelay, 4) >= 0);

    LOG("Connected to %s:%s\n", host, port);
    writeFd = readFd;
}

int Stream::getchar() {
    char c;
    int r = read(readFd, &c, 1);
    return r ? c : EOF;
}

void Stream::send(const string& msg) {
    write(writeFd, msg.data(), msg.size());
}

bool Stream::readable() {
    fd_set fdset;
    FD_ZERO(&fdset);
    FD_SET(readFd, &fdset);
    struct timeval tv;
    tv.tv_sec = 0;
    tv.tv_usec = 1000;
    int r = select(readFd+1, &fdset, NULL, NULL, &tv);
    CHECK(r >= 0);
    return r;
}

void split(const string& s, vector<string>* v) {
    string c;
    for (int i = 0; i < s.size(); i++) {
        if (s[i] == ' ') {
            v->push_back(c);
            c.clear();
        }
        else {
            c += s[i];
        }
    }
}

float parse_float(const string& s) {
    float f = 0.0;
    float r = 0.0;
    int i = 0;
    if (s[0] == '-') i++;
    for (; i < s.size(); i++) {
        if (s[i] == '.') {
            r = 0.1;
        }
        else {
            int v = s[i] - '0';
            CHECKL(v >= 0 && v <= 9, "%d %c\n", v, s[i]);
            if (r == 0.0) {
                f = f * 10 + v;
            }
            else {
                f += r * v;
                r *= 0.1;
            }
        }
    }
    if (s[0] == '-') {
        f = -f;
    }
    return f;
}

float parse_float2(const string& s) {
    return parse_float(s);
}

int parse_int(const string& s) {
    int r = 0;
    int i = 0;
    if (s[0] == '-') i++;
    for (; i < s.size(); i++) {
        int v = s[i] - '0';
        CHECKL(v >= 0 && v <= 9, "%d\n", v);
        r = r * 10 + v;
    }
    if (s[0] == '-') {
        r = -r;
    }
    return r;
}

void parse_line(Stream* stream, Rover* rover) {
    string line;
    while (true) {
        int c = stream->getchar();
        if (c == ';') {
            break;
        }
        else if (c == '\n') {
            // for debugging
            continue;
        }
        if (c == EOF) {
            LOG("Done\n");
            done = true;
            return;
        }
        line += c;
    }
    //LOG("recv: %s\n", line.c_str());

    vector<string> v;
    split(line, &v);
    CHECK(!v.empty());
    CHECKL(v[0].size() == 1, "%s\n", v[0].c_str());
    switch (line[0]) {
    case 'I': {
        CHECK(v.size() == 9);
        float dx = parse_float(v[1]);
        float dy = parse_float(v[2]);
        int time_limit = parse_int(v[3]);
        float min_sensor = parse_float(v[4]);
        float max_sensor = parse_float(v[5]);
        float max_speed = parse_float(v[6]);
        float max_turn = parse_float2(v[7]);
        float max_hard_turn = parse_float2(v[8]);

        rover->init(dx,
                    dy,
                    time_limit,
                    min_sensor,
                    max_sensor,
                    max_speed,
                    max_turn,
                    max_hard_turn);
        break;
    }
    case 'T': {
        int timestamp = parse_int(v[1]);
        char accState = v[2][0];
        char turnState = v[2][1];
        float x = parse_float(v[3]);
        float y = parse_float(v[4]);
        float dir = parse_float2(v[5]);
        float speed = parse_float(v[6]);
        rover->setState(timestamp, accState, turnState, x, y, dir, speed);

        rover->clearObjects();
        for (int i = 7; i < v.size(); i++) {
            CHECK(v.size() > i + 3);
            CHECK(v[i].size() == 1);
            char type = v[i][0];
            float x = parse_float(v[++i]);
            float y = parse_float(v[++i]);
            if (type == 'm') {
                float dir = parse_float2(v[++i]);
                float speed = parse_float(v[++i]);
                rover->addObject(new Obj(x, y, dir, speed));
            }
            else {
                float r = parse_float(v[++i]);
                rover->addObject(new Obj(type, x, y, r));
            }
        }
        come_tel = true;

        break;
    }
    case 'B': {
        int timestamp = parse_int(v[1]);
        rover->reportCrash(timestamp);
        break;
    }
    case 'C':
    case 'K': {
        int timestamp = parse_int(v[1]);
        rover->reportFail(line[0], timestamp);
        break;
    }
    case 'S': {
        int timestamp = parse_int(v[1]);
        rover->reportSuccess(timestamp);
        break;
    }
    case 'E': {
        int timestamp = parse_int(v[1]);
        rover->reportDone(timestamp);
        break;
    }
    default:
        CHECK(false);
    }
}

void parse(Stream* stream, Rover* rover) {
    come_tel = false;
    while (stream->readable()) {
        parse_line(stream, rover);
        if (done || come_tel) break;
    }
}
