#include <math.h>
#include <time.h>
#include <sys/time.h>

#include <pthread.h>

#include <google/gflags.h>

#include "rover.h"
#include "log.h"
#include "parser.h"
#include "util.h"
#include "geom.h"
#include "world.h"

DEFINE_double(danger_zone, 0.75, "");
//DEFINE_double(escape_distance, 1.0, "");
//DEFINE_double(escape_distance_c, 2.0, "");
DEFINE_double(escape_distance, 0.2, "");
DEFINE_double(escape_distance_c, 0.5, "");
DEFINE_int32(latency, 100, "");

DEFINE_double(size_endo, 2.0, "");

DEFINE_double(too_danger_ratio, 0.3, "");
DEFINE_double(too_danger_distance, 1, "");
DEFINE_double(too_danger_angle, 0.51 * M_PI, "");

DEFINE_double(near_goal, 1, "");

DEFINE_int32(verbose, 0, "");

DEFINE_double(need_adjustment_distance, 10, "");

DEFINE_double(adjust_times, 20, "");
DEFINE_double(adjust_delta, 0.003, "");

DEFINE_double(simulation_delta, 0.01, "");
//DEFINE_double(simulation_times, 100, "");
DEFINE_double(simulation_times, 100, "");
//DEFINE_double(simulation_times, 25, "");

//DEFINE_double(simulation_delta, 0.05, "");
//DEFINE_double(simulation_times, 20, "");

Obj::Obj(char type0, float x0, float y0, float r0)
    : type(type0), x(x0), y(y0), r(r0), dir(0), speed(0) {
}
Obj::Obj(float x0, float y0, float dir0, float speed0)
    : type('m'), x(x0), y(y0), r(FLAGS_size_endo), dir(dir0), speed(speed0) {
}

long long getTime() {
    struct timeval tv;
    CHECKL(gettimeofday(&tv, 0) >= 0, "gettimeofday\n");
    return 1000LL * tv.tv_sec + tv.tv_usec / 1000;
}

static int rightState(char turnState) {
    if (turnState == 'L') {
        turnState = 'l';
    }
    else if (turnState == 'l') {
        turnState = '-';
    }
    else if (turnState == '-') {
        turnState = 'r';
    }
    else if (turnState == 'r') {
        turnState = 'R';
    }
    return turnState;
}

static int leftState(char turnState) {
    if (turnState == 'R') {
        turnState = 'r';
    }
    else if (turnState == 'r') {
        turnState = '-';
    }
    else if (turnState == '-') {
        turnState = 'l';
    }
    else if (turnState == 'l') {
        turnState = 'L';
    }
    return turnState;
}

Rover::Rover(Stream* stream0)
    : stream(stream0), inited(false), firstTel(false), goal(0),
      world(new World(this)) {
    rotAccelFirstGuess = true;
    guessedRotAcc = dtor(120);
    reset();
}

void Rover::init(float dx,
                 float dy,
                 int time_limit,
                 float min_sensor,
                 float max_sensor,
                 float max_speed,
                 float max_turn,
                 float max_hard_turn) {
    CHECK(!inited);
    inited = true;
    this->dx = dx;
    this->dy = dy;
    this->time_limit = time_limit;
    this->min_sensor = min_sensor;
    this->max_sensor = max_sensor;
    this->max_speed = max_speed;
    this->max_turn = dtor(max_turn);
    this->max_hard_turn = dtor(max_hard_turn);

    LOG("Init: x=%.3f y=%.3f time_limit=%d "
        "sensor=(%.3f, %.3f) max_speed=%.3f max_turn=%.1f max_hard_turn=%.1f\n",
        dx, dy, time_limit, min_sensor, max_sensor,
        max_speed, max_turn, max_hard_turn);

    reset();
    //mode = SEARCH;
}

void Rover::reset() {
    firstTel = false;
    state_time = -1;
    guessedAccel = 2.0;
    prevDirChange = -1;

    canGuessAccel = false;
    canGuessRotAccel = false;
    accState = 'a';
    turnState = '-';
    prev_cmd_time = 0;
    should_brake = false;
    prev_right = false;
    prev_left = false;

    gx = gy = 0;

    speed = 0;
    dir = reported_dir = 0;
    rot = 0;

    world->reset();
    avoiding_emergency = false;
    will_die = false;
}

void Rover::guessAccel(int new_time, float new_speed) {
    if (!canGuessAccel) {
        return;
    }
    for (int i = 0; i < emittedCmds.size(); i++) {
        if (emittedCmds[i].second.find('a') == string::npos) {
            return;
        }
    }
    if (new_time == state_time) return;
    float nga = (new_speed - prev_speed) / (new_time - state_time) * 1000.0;
    if (nga > 0) {
        guessedAccel = nga;
    }
    LOG("guessed accel: %f\n", guessedAccel);
}

bool Rover::guessRotAccel(int t, float d, float ts) {
    if (!canGuessRotAccel) return false;
    int td = t - state_time;
    if (td <= 0) return false;
    if (ts == '-' || reported_turn_state == '-') return false;
    if (ts != reported_turn_state) return false;

    float dirChange = fabs((d - reported_dir) / td * 1000);
    if (ts == 'R' && dirChange < max_hard_turn ||
        ts == 'r' && dirChange < max_turn ||
        ts == 'l' && dirChange < max_turn ||
        ts == 'L' && dirChange < max_hard_turn) {
        if (prevDirChange > 0) {
            float gra = fabs((dirChange - prevDirChange) / td * 1000);
            if (rotAccelFirstGuess) {
                rotAccelFirstGuess = false;
                guessedRotAcc = gra;
            }
            else {
                guessedRotAcc = max(guessedRotAcc, gra);
            }
            LOG("guessed rot accel: %f %f=>%f\n",
                rtod(guessedRotAcc), rtod(prevDirChange), rtod(dirChange));
        }

        prevDirChange = dirChange;
        return true;
    }
    else {
        return false;
    }
}

void Rover::setState(int timestamp, char accState, char turnState,
                     float x, float y, float dir, float speed) {
    if (canGuessRotAccel) {
        if (timestamp > state_time) {
            int td = timestamp - state_time;
            rot = diffDir(dtor(dir), reported_dir) / td * 1000;
        }
    }

    //guessPos((timestamp - prev_cmd_time) * 0.001);
    LOG("Tel: state=%c%c t=%d x=%.3f y=%.3f d=%.1f s=%.3f rot=%.3f\n",
        accState, turnState, timestamp, x, y, dir, speed, rtod(rot));

    guessAccel(timestamp, speed);
    if (!guessRotAccel(timestamp, dtor(dir), turnState)) {
        prevDirChange = -1;
    }

    this->accState = accState;
    reported_turn_state = this->turnState = turnState;
    this->x = x;
    this->y = y;
    reported_dir = this->dir = dtor(dir);
    prev_speed = this->speed = speed;

    state_time = timestamp;
    state_real_time = getTime();

    firstTel = true;

    prev_right = true;
    prev_left = true;
    for (int i = 0; i < emittedCmds.size(); i++) {
        if (emittedCmds[i].second.find('r') == string::npos) {
            prev_right = false;
        }
        else if (emittedCmds[i].second.find('l') == string::npos) {
            prev_left = false;
        }
    }

    replayCmds();
    emittedCmds.clear();
    canGuessAccel = timestamp;
    canGuessRotAccel = timestamp;

    should_brake = false;

    //world->added = true;
    world->calcPath(x, y, this->dir);
    //world->calcPath(x, y, this->reported_dir);
}

void Rover::replayCmds() {
    int t = state_time;
    char as = accState;
    char ts = turnState;
    for (int i = 0; i <= emittedCmds.size(); i++) {
        int nt;
        if (i == emittedCmds.size()) {
            nt = state_time + FLAGS_latency;
        }
        else {
            nt = emittedCmds[i].first;
        }
        float elapsed = 0.001 * (nt - t);
        x += cos(dir) * speed * elapsed;
        y += sin(dir) * speed * elapsed;
#if 0
        if (ts == 'R') {
            dir -= elapsed * max_hard_turn;
            //dir -= elapsed * (prev_right ? max_hard_turn : max_turn);
        }
        else if (ts == 'r') {
            dir -= max_turn * elapsed;
        }
        else if (ts == 'l') {
            dir += max_turn * elapsed;
        }
        else if (ts == 'L') {
            //LOG("%f => %f\n", rtod(dir), rtod(dir + max_hard_turn * elapsed));
            dir += elapsed * max_hard_turn;
            //dir += elapsed * (prev_left ? max_hard_turn : max_turn);
        }
#else
        float orig_rot = rot;
        simulateDir(elapsed, ts, &rot, &dir);
        LOG2("rot change: %f=>%f (%.3f sec)\n",
             rtod(orig_rot), rtod(rot), elapsed);
#endif
        if (as == 'a') {
            speed += guessedAccel * elapsed;
        }
        else if (as == 'b') {
            // TODO: rough estimation
            speed -= guessedAccel * elapsed;
        }

        if (i == emittedCmds.size()) {
            break;
        }

        t = nt;
        const string& cmd = emittedCmds[i].second;
        if (cmd.find('a') != string::npos) {
            as = 'a';
        }
        if (cmd.find('r') != string::npos) {
            ts = rightState(ts);
        }
        else if (cmd.find('l') != string::npos) {
            ts = leftState(ts);
        }
    }

    accState = as;
    turnState = ts;

    LOG("Replayed: state=%c%c t=%d x=%.3f y=%.3f d=%.1f s=%.3f rot=%.3f\n",
        accState, turnState, state_time + FLAGS_latency,
        x, y, rtod(dir), speed, rtod(rot));
    /*
    this->x += this->speed * cos(this->dir);
    this->y += this->speed * sin(this->dir);
    */
}

void Rover::reportCrash(int timestamp) {
    CHECK(inited);

    LOG("Crash: %d\n", timestamp);
    canGuessAccel = false;
    canGuessRotAccel = false;
    rot = 0;

    sortObject();
    for (int i = 0; i < objs.size(); i++) {
        Obj* o = objs[i];
        if (o->type != 'c') break;
        if (o->distance < 10) {
            LOG("brake!\n");
            should_brake = true;
            break;
        }
    }
}

void Rover::reportFail(char why, int timestamp) {
    CHECK(inited);

    LOG("Fail(%c): %d\n", why, timestamp);
}

void Rover::reportSuccess(int timestamp) {
    CHECK(inited);

    LOG("Success: %d\n", timestamp);
}

void Rover::reportDone(int timestamp) {
    CHECK(inited);
    //inited = false;
    reset();

    LOG("Done: %d\n", timestamp);
}

void Rover::clearObjects() {
    CHECK(inited);
    for (int i = 0; i < objs.size(); i++) {
        delete objs[i];
    }
    objs.clear();
}

void Rover::addObject(Obj* obj) {
    LOG("Obj: %c %.3f %.3f %.3f\n", obj->type, obj->x, obj->y, obj->r);

    CHECK(inited);
    switch (obj->type) {
    case 'h': {
        if (goal) {
            delete obj;
        }
        else {
            goal = obj;
            //mode = GOAL;
        }
        break;
    }
    case 'm':
        obj->x += cos(obj->dir) * obj->speed * FLAGS_latency / 1000;
        obj->y += sin(obj->dir) * obj->speed * FLAGS_latency / 1000;
    default:
        objs.push_back(obj);
        break;
    }

    if (obj->type == 'b' || obj->type == 'c') {
        world->addObject(obj);
    }
}

void Rover::sortObject() {
    for (int i = 0; i < objs.size(); i++) {
        Obj* o = objs[i];
        o->distance = distance(o->x, o->y, x, y) - o->r;
    }

    struct SortObj {
        bool operator()(const Obj* o1, const Obj* o2) {
            // 'c'&3 => 3
            // 'b'&3 => 2
            // 'm'&3 => 1
            int t1 = o1->type & 3;
            int t2 = o2->type & 3;
            if (t1 > t2) return true;
            if (t1 < t2) return false;
            if (o1->distance < o2->distance) return true;
            if (o1->distance > o2->distance) return false;
            return o1 < o2;
        }
    };

    for (int i = 0; i < objs.size(); i++) {
        Obj* o = objs[i];
        //LOG("Obj: %c %.3f %.3f %.3f\n", o->type, o->x, o->y, o->r);
    }
}

void Rover::turnR() {
    turnAttempt = 'r';
    turnState = rightState(turnState);
}

void Rover::turnL() {
    turnAttempt = 'l';
    turnState = leftState(turnState);
}

bool Rover::danger() {
    float gd = distance2(gx, gy, x, y);
    for (int i = 0; i < objs.size(); i++) {
        Obj* o = objs[i];
        if (o->type != 'c') break;
        float d = distance2(o->x, o->y, x, y);
        if (d < 100 && d > gd) {
            return true;
        }
    }
    return false;
}

bool Rover::nearGoal() {
    float goalDistance = length(x, y);
    return goalDistance < FLAGS_near_goal * max_speed;
}

static int turnStateNum(char ts) {
    switch (ts) {
    case 'L': return -2;
    case 'l': return -1;
    case '-': return 0;
    case 'r': return 1;
    case 'R': return 2;
    }
    CHECKL(false, "%c\n", ts);
}

void Rover::gotoXY(float gx, float gy) {
    float r = atan2(gy-y, gx-x) - dir;
    r = normalizeDir(r);

    bool allowR = true;
    bool allowL = true;
#if 0
    for (int i = 0; i < objs.size(); i++) {
        Obj* o = objs[i];
        if (o->type != 'c') {
            continue;
        }

        float d = distance(o->x, o->y, x, y) - o->r - 0.5;
        float danger_distance = max(FLAGS_too_danger_ratio * o->r,
                                    FLAGS_too_danger_distance);
        if (d < danger_distance) {
            float r = diffDir(atan2(o->y - y, o->x - x), dir);
            if (r > FLAGS_too_danger_angle) {
            }
            else if (r > 0) {
                allowL = false;
                LOG("allowL=false\n");
            }
            else if (r > - FLAGS_too_danger_angle) {
                allowR = false;
                LOG("allowR=false\n");
            }
        }
    }
    //allowL = true;
    //allowR = true;
#endif

    char origTurnState = turnState;
    turnAttempt = '-';
    char accAttempt = 'a';

    char desiredState = '-';

    int times = 1;

    LOG("%.3f %.3f r=%.1f\n", gx, gy, rtod(r));

    float goalDistance = distance(gx, gy, x, y);

#if 0
    if (!allowL && allowR && (turnState == 'L' || turnState == 'l')) {
        turnR();
    }
    else if (!allowR && allowL && (turnState == 'R' || turnState == 'r')) {
        turnL();
    }
    else
#endif
    if (fabs(r) < 0.01) {
    }
    //else if (r > max_hard_turn * 0.1 && allowL) {
    //else if (r > max_hard_turn * 0.1 && r > rot * 3) {
    else if (r > max_hard_turn * 0.1 && goalDistance > 5) {
        turnAttempt = 'l';
        turnState = leftState(turnState);
        //times = 4;

        desiredState = 'L';
        //turnL();

/*
        if (r > M_PI * 0.6 && state_time > 10000 &&
            gx == 0 && gy == 0) {
            LOG("Braking for opposite goal\n");
            accAttempt = 'b';
        }
*/
    }
    //else if (r < -max_hard_turn * 0.1 && allowR) {
    //else if (r < -max_hard_turn * 0.1 && allowR && r < rot * 3) {
    else if (r < -max_hard_turn * 0.1 && goalDistance > 5) {
        turnAttempt = 'r';
        turnState = rightState(turnState);
        //times = 4;

        desiredState = 'R';
        //turnR();

/*
        if (r < -M_PI * 0.6 && state_time > 10000 &&
            gx == 0 && gy == 0) {
            LOG("Braking for opposite goal\n");
            accAttempt = 'b';
        }
*/
    }
    else if (r > 0) {
        if (turnState == 'L') {
            turnAttempt = 'r';
            turnState = rightState(turnState);
            //turnR();
        }
        else if (turnState == 'l') {
        }
        else {
            turnAttempt = 'l';
            turnState = leftState(turnState);
            //if (allowL) turnL();
        }

        desiredState = 'l';
    }
    else {
        if (turnState == 'R') {
            turnAttempt = 'l';
            turnState = leftState(turnState);
            //turnL();
        }
        else if (turnState == 'r') {
        }
        else {
            turnAttempt = 'r';
            turnState = rightState(turnState);
            //if (allowR) turnR();
        }

        desiredState = 'r';
    }

    if (accAttempt == 'a' &&
        world->needBrake(x, y, speed)) {
        LOG("Braking for sharp turn\n");
        accAttempt = 'b';
    }

    char origAcc = accAttempt;
    char origTurn = turnAttempt;
    avoiding_emergency = false;
    will_die = false;

#if 1
    char origDesiredState = desiredState;
    if (avoidEmergency2(origTurnState, &accAttempt, &desiredState)) {
        // Consider plan again!
        world->added = true;

        LOG("avoid DIE? %c%c => %c%c\n",
            origAcc, origDesiredState, accAttempt, desiredState);
        avoiding_emergency = true;

        int tn = turnStateNum(origTurnState);
        int dn = turnStateNum(desiredState);
        //while (tn != dn) {
        if (tn != dn) {
            string ctrl;
            ctrl += accAttempt;

            if (tn > dn) {
                ctrl += 'l';
                tn--;
            }
            else {
                ctrl += 'r';
                tn++;
            }

            ctrl += ';';
            send(ctrl, false);
        }
        accState = accAttempt;
        turnState = desiredState;
        return;
    }
#else
    if (avoidEmergency(&accAttempt, &turnAttempt)) {
        // Consider plan again!
        world->added = true;

        LOG("avoid DIE? %c%c => %c%c\n",
            origAcc, origTurn, accAttempt, turnAttempt);
        avoiding_emergency = true;
        //times = 4;
    }
#endif

    for (int i = 0; i < times; i++) {
        //string ctrl = "a";
        string ctrl;
        ctrl += accAttempt;
        if (turnAttempt != '-') {
            ctrl += turnAttempt;
        }
        ctrl += ';';
        send(ctrl, origTurnState == turnState);
    }
}

int Rover::simulateSafety(float sim_x, float sim_y,
                          float sim_speed, float sim_dir, float sim_rot,
                          float accAtt, float turnAtt,
                          const vector<Obj*>& nearObjs) {
    for (int i = 0; i < FLAGS_simulation_times; i++) {
        sim_x += cos(sim_dir) * sim_speed * FLAGS_simulation_delta;
        sim_y += sin(sim_dir) * sim_speed * FLAGS_simulation_delta;

        if (length(sim_x, sim_y) < 5) {
            // we're back!
            //return true;
            return 0;
        }

        if (accAtt == 'a') {
            sim_speed += guessedAccel * FLAGS_simulation_delta;
        }
        else if (accAtt == 'b') {
            sim_speed -= guessedAccel * FLAGS_simulation_delta;
        }
#if 1
        if (turnAtt == 'l') {
            simulateDir(FLAGS_simulation_delta, 'L', &sim_rot, &sim_dir);
        }
        else if (turnAtt == 'r') {
            simulateDir(FLAGS_simulation_delta, 'R', &sim_rot, &sim_dir);
        }
        else {
            simulateDir(FLAGS_simulation_delta, '-', &sim_rot, &sim_dir);
        }
#else
        if (turnAtt == 'l') {
            sim_dir += max_turn * FLAGS_simulation_delta;
        }
        else if (turnAtt == 'r') {
            sim_dir -= max_turn * FLAGS_simulation_delta;
        }
#endif

        for (int j = 0; j < nearObjs.size(); j++) {
            const Obj* o = nearObjs[j];
            if (distance(o->x, o->y, sim_x, sim_y) - 0.5 - o->r < 0) {
                //return false;
                return i + 1;
            }
        }
    }
    return 0;
}

int Rover::simulateSafety2(float sim_x, float sim_y,
                           float sim_speed, float sim_dir, float sim_rot,
                           char origTurnState,
                           char accAtt, char turnAtt,
                           const vector<Obj*>& nearObjs) {
    float accTime = 0;
    for (int i = 0; i < FLAGS_simulation_times; i++) {
        accTime += FLAGS_simulation_delta;

        sim_x += cos(sim_dir) * sim_speed * FLAGS_simulation_delta;
        sim_y += sin(sim_dir) * sim_speed * FLAGS_simulation_delta;

        if (length(sim_x, sim_y) < 5) {
            // we're back!
            return 0;
        }

        if (accAtt == 'a') {
            sim_speed += guessedAccel * FLAGS_simulation_delta;
        }
        else if (accAtt == 'b') {
            sim_speed -= guessedAccel * FLAGS_simulation_delta;
        }
        // There may be short delay...
        if (accTime < 0.05) {
            simulateDir(FLAGS_simulation_delta, origTurnState,
                        &sim_rot, &sim_dir);
        }
        else {
            simulateDir(FLAGS_simulation_delta, turnAtt, &sim_rot, &sim_dir);
        }

        for (int j = 0; j < nearObjs.size(); j++) {
            const Obj* o = nearObjs[j];
            //if (distance(o->x, o->y, sim_x, sim_y) - 0.5 - o->r < 0) {
            if (distance(o->x, o->y, sim_x, sim_y) - o->r < 0) {
                LOG2("Finished... %c%c %d @(%.3f %.3f) d=%.1f r=%f\n",
                     accAtt, turnAtt, i+1, sim_x, sim_y,
                     rtod(sim_dir), rtod(sim_rot));
                return i + 1;
            }
        }
    }
    return 0;
}

bool Rover::avoidEmergency2(char origTurnState,
                            char* accAtt,
                            char* turnAtt) {
    vector<Obj*> nearObjs;
    for (int i = 0; i < objs.size(); i++) {
        Obj* o = objs[i];
        if ((o->type == 'c' || o->type == 'm') &&
            o->distance - 0.5 < max_speed) {
            nearObjs.push_back(o);
        }
    }

    if (nearObjs.empty()) {
        return false;
    }

    int bestLife = simulateSafety2(x, y, speed, dir, rot,
                                   origTurnState, *accAtt, *turnAtt, nearObjs);
    if (bestLife) {
        char bestAcc = *accAtt;
        char bestTurn = *turnAtt;
        bool retried = false;

        LOG("May DIE!: %d\n", state_time);

        char a = *accAtt;

        const char* trials = 0;
        switch (*turnAtt) {
        case 'R':
            trials = "r-lL";
            break;
        case 'r':
            trials = "-RlL";
            break;
        case '-':
            trials = "lrRL";
            break;
        case 'l':
            trials = "-LrR";
            break;
        case 'L':
            trials = "l-rR";
            break;
        default:
            CHECKL(false, "%c\n", *turnAtt);
            return false;
        }

    retry:
        for (int i = 0; i < 4; i++) {
            int life = simulateSafety2(x, y, speed, dir, rot,
                                       origTurnState, a, trials[i], nearObjs);
            if (!life) {
                *accAtt = a;
                *turnAtt = trials[i];
                return true;
            }
            LOG2("%c%c: %d\n", a, trials[i], life);
            if (life > bestLife) {
                bestLife = life;
                bestAcc = a;
                bestTurn = trials[i];
            }
        }
        if (!retried) {
            retried = true;
            if (a == 'a') {
                a = 'b';
            }
            else if (a == 'b') {
                a = 'a';
            }
            goto retry;
        }
        LOG("Definitely will DIE!: %d %c%c => %c%c %d\n",
            state_time, *accAtt, *turnAtt, bestAcc, bestTurn, bestLife);
        will_die = true;

        if (bestAcc != *accAtt || bestTurn != *turnAtt) {
            *accAtt = bestAcc;
            *turnAtt = bestTurn;
            return true;
        }
    }
    return false;
}

bool Rover::avoidEmergency(char* accAtt, char* turnAtt) {
    vector<Obj*> nearObjs;
    for (int i = 0; i < objs.size(); i++) {
        Obj* o = objs[i];
        if ((o->type == 'c' || o->type == 'm') &&
            o->distance - 0.5 < max_speed) {
            nearObjs.push_back(o);
        }
    }

    if (nearObjs.empty()) {
        return false;
    }

    int bestLife =
        simulateSafety(x, y, speed, dir, rot, *accAtt, *turnAtt, nearObjs);
    if (bestLife) {
        char bestAcc = *accAtt;
        char bestTurn = *turnAtt;

        LOG("May DIE!: %d\n", state_time);

        char a = *accAtt;

#define UPDATE_LIFE(ch)                         \
        do {                                    \
            if (life > bestLife) {              \
                bestLife = life;                \
                bestAcc = a;                    \
                bestTurn = ch;                  \
            }                                   \
        } while(0)                              \

    retry:
        if (*turnAtt == 'r') {
            int life = simulateSafety(x, y, speed, dir, rot, a, '-', nearObjs);
            if (!life) {
                *accAtt = a;
                *turnAtt = '-';
                return true;
            }
            UPDATE_LIFE('-');
            life = simulateSafety(x, y, speed, dir, rot, a, 'l', nearObjs);
            if (!life) {
                *accAtt = a;
                *turnAtt = 'l';
                return true;
            }
            UPDATE_LIFE('l');
        }
        else if (*turnAtt == 'l') {
            int life = simulateSafety(x, y, speed, dir, rot, a, '-', nearObjs);
            if (!life) {
                *accAtt = a;
                *turnAtt = '-';
                return true;
            }
            UPDATE_LIFE('-');
            life = simulateSafety(x, y, speed, dir, rot, a, 'r', nearObjs);
            if (!life) {
                *accAtt = a;
                *turnAtt = 'r';
                return true;
            }
            UPDATE_LIFE('r');
        }
        else if (*turnAtt == '-') {
            int life = simulateSafety(x, y, speed, dir, rot, a, 'l', nearObjs);
            if (!life) {
                *accAtt = a;
                *turnAtt = 'l';
                return true;
            }
            UPDATE_LIFE('l');
            life = simulateSafety(x, y, speed, dir, rot, a, 'r', nearObjs);
            if (!life) {
                *accAtt = a;
                *turnAtt = 'r';
                return true;
            }
            UPDATE_LIFE('r');
        }
        if (a == 'a') {
            a = 'b';
            goto retry;
        }
        LOG("Definitely will DIE!: %d %c%c => %c%c %d\n",
            state_time, *accAtt, *turnAtt, bestAcc, bestTurn, bestLife);
        will_die = true;

        if (bestAcc != *accAtt || bestTurn != *turnAtt) {
            *accAtt = bestAcc;
            *turnAtt = bestTurn;
            return true;
        }
    }
    return false;
}

void Rover::updateTarget() {
    targetX = rnd(dx) - dx / 2;
    targetY = rnd(dy) - dy / 2;
    LOG("target=(%.3f, %.3f)\n", targetX, targetY);
}

void Rover::calcBestGoal(float* gx_out, float* gy_out) {
    float gx = 0;
    float gy = 0;

    float dx = gx - x;
    float dy = gy - y;
    float a = dy / dx;
    float b = gy - gx * a;

    //HalfLine hl(gx, gy, x, y);
    HalfLine hl(dir, x, y);

    //float danger_zone = FLAGS_danger_zone * speed;
    float danger_zone = 1.0;
    float escape_distance = FLAGS_escape_distance * speed;
    float escape_distance_c = FLAGS_escape_distance_c * speed;
    LOG2("dz=%.3f ed=%.3f edc=%.3f\n",
         danger_zone, escape_distance, escape_distance_c);

    for (int i = 0; i < objs.size(); i++) {
        Obj* o = objs[i];

        float d = distance(o->x, o->y, x, y) - o->r - 0.5;
        if (d > 2) {
            float ax1 = o->x + cos(hl.dir + M_PI / 2) * (o->r + 0.5);
            float ay1 = o->y + sin(hl.dir + M_PI / 2) * (o->r + 0.5);
            //float d1 = distance2(x, y, ax1, ay1);
            float dr1 = fabs(diffDir(atan2(ay1 - y, ax1 - x), dir));
            float ax2 = o->x + cos(hl.dir - M_PI / 2) * (o->r + 0.5);
            float ay2 = o->y + sin(hl.dir - M_PI / 2) * (o->r + 0.5);
            //float d2 = distance2(x, y, ax2, ay2);
            float dr2 = fabs(diffDir(atan2(ay2 - y, ax2 - x), dir));

            float dr3 = fabs(diffDir(atan2(o->y - y, o->x - x), dir));

            float dd = max(dr1, dr2);
            dd = min(dr3, dd);

            if (dd > 1) {
                LOG("ignore obj distance: %c (%.3f, %.3f) r=%.3f d=%.3f\n",
                    o->type, o->x, o->y, o->r, d);
                continue;
            }
        }

        if (o->type == 'c' || o->type == 'm') {
            if (d > escape_distance_c) {
                LOG("ignore obj distance: %c (%.3f, %.3f) r=%.3f d=%.3f\n",
                    o->type, o->x, o->y, o->r, d);
                continue;
            }
        }
        else {
            if (d > escape_distance) {
                continue;
            }
        }

        d = hl.distance(o->x, o->y);
        LOG("obj distance: %c (%.3f, %.3f) r=%.3f d=%.3f\n",
            o->type, o->x, o->y, o->r, d);
        if (o->r + danger_zone < d) continue;

        float ax1 = o->x + cos(hl.dir + M_PI / 2) * (o->r + danger_zone);
        float ay1 = o->y + sin(hl.dir + M_PI / 2) * (o->r + danger_zone);
        //float d1 = distance2(x, y, ax1, ay1);
        float dr1 = diffDir(atan2(ay1 - y, ax1 - x), dir);
        float ax2 = o->x + cos(hl.dir - M_PI / 2) * (o->r + danger_zone);
        float ay2 = o->y + sin(hl.dir - M_PI / 2) * (o->r + danger_zone);
        //float d2 = distance2(x, y, ax2, ay2);
        float dr2 = diffDir(atan2(ay2 - y, ax2 - x), dir);

        float cd = 0;
        float cd2 = 0;
        if (fabs(dr1) < fabs(dr2)) {
            gx = ax1;
            gy = ay1;
            cd = dr1;
            cd2 = dr2;
        }
        else {
            gx = ax2;
            gy = ay2;
            cd = dr2;
            cd2 = dr1;
        }
        LOG("escape from: %c (%.3f, %.3f) %.3f (%.3f)\n",
            o->type, gx, gy, cd, cd2);
        break;
    }

    *gx_out = gx;
    *gy_out = gy;
}

int Rover::currentTime() {
    int time = 0;
    if (state_time >= 0) {
        time = state_time + FLAGS_latency + getTime() - state_real_time;
    }
    return time;
}

void Rover::send(const string& ctrl, bool suppressDup) {
    char origAccState = accState;
    if (ctrl.find('a') != string::npos) {
        if (accState == '-') {
            accState = 'a';
        }
        else if (accState == 'b') {
            accState = '-';
        }
    }
    else if (ctrl.find('b') != string::npos) {
        if (accState == '-') {
            accState = 'b';
        }
        else if (accState == 'a') {
            accState = '-';
        }
    }

    if (accState != origAccState) {
        suppressDup = false;
    }

    int time = currentTime();
    prev_cmd_time = time;
    prev_cmd = ctrl;

    //LOG("attempt send: %s %d\n", ctrl.c_str(), emittedCmds.size());

    if (!emittedCmds.empty() && emittedCmds.back().second == ctrl) {
        if (suppressDup) {
            return;
        }
/*
        int i = ctrl.find('r');
        if (i != string::npos) {

        }
        else if ((i = ctrl.find('l')) != string::npos) {
        }
        else {
            return;
        }
*/
    }

    emittedCmds.push_back(make_pair(time, ctrl));
    LOG("Send: %s %d\n", ctrl.c_str(), time);
    stream->send(ctrl);
}

void Rover::simulateDir(float elapsed, char ts,
                        float* rot_out, float* dir_out) {
    float rot = *rot_out;
    float dir = *dir_out;

    dir += rot * elapsed;
    if (ts == 'R') {
        rot = max(rot - guessedRotAcc * elapsed, -max_hard_turn);
    }
    else if (ts == 'r') {
        if (rot > -max_turn) {
            rot = max(rot - guessedRotAcc * elapsed, -max_turn);
        }
        else {
            rot = min(rot + guessedRotAcc * elapsed, -max_turn);
        }
    }
    else if (ts == '-') {
        if (rot < 0) {
            rot = min(rot + guessedRotAcc * elapsed, 0.0f);
        }
        else {
            rot = max(rot - guessedRotAcc * elapsed, 0.0f);
        }
    }
    else if (ts == 'l') {
        if (rot < max_turn) {
            rot = min(rot + guessedRotAcc * elapsed, max_turn);
        }
        else {
            rot = max(rot - guessedRotAcc * elapsed, max_turn);
        }
    }
    else if (ts == 'L') {
        rot = min(rot + guessedRotAcc * elapsed, max_hard_turn);
    }

    *rot_out = rot;
    *dir_out = dir;
}

void Rover::guessPos(float elapsed) {
    if (elapsed == 0.0) {
        elapsed = (currentTime() - prev_cmd_time) * 0.001;
    }

    x += cos(dir) * speed * elapsed;
    y += sin(dir) * speed * elapsed;
    simulateDir(elapsed, turnState, &rot, &dir);
    if (accState == 'a') {
        speed += guessedAccel * elapsed;
    }
    else if (accState == 'b') {
        speed -= guessedAccel * elapsed;
    }

    LOG("Guessed: (e=%.3f) state=%c%c x=%.3f y=%.3f d=%.1f s=%.3f\n",
        elapsed, accState, turnState, x, y, rtod(dir), speed);

    for (int i = 0; i < objs.size(); i++) {
        Obj* o = objs[i];
        if (o->type == 'm') {
            o->x += cos(o->dir) * o->speed * elapsed;
            o->y += sin(o->dir) * o->speed * elapsed;
        }
    }
}

bool Rover::checkNearGoalMissing() {
    return false;

    if (nearGoal() &&
        (objs.empty() || objs.front()->type != 'm' ||
         objs.front()->distance > FLAGS_near_goal * max_speed * 2)) {
        float dd = diffDir(atan2(-y, -x), dir);
        if (dd > M_PI * 0.3) {
            LOG("near goal!\n");
            canGuessAccel = false;
            send("bl;");
            return true;
        }
        else if (dd < -M_PI * 0.3) {
            LOG("near goal!\n");
            canGuessAccel = false;
            send("br;");
            return true;
        }
    }
    return false;
}

void Rover::adjustNearOverlap() {
    float gd = distance(gx, gy, x, y);

    for (int i = 0; i < objs.size(); i++) {
        Obj* o = objs[i];

        if (o->type == 'm') continue;

        float d = distance(o->x, o->y, x, y);
        if (gd < d) continue;
        d -= o->r + 0.5;
        if (d > FLAGS_need_adjustment_distance) {
            continue;
        }

        HalfLine hl(gx, gy, x, y);
        float hd = hl.distance(o->x, o->y);
        if (hd < o->r + 0.5) {
            float d = hl.dir;
            float pd = hd;
            for (int i = 0; i < FLAGS_adjust_times; i++) {
                HalfLine hl2(d + FLAGS_adjust_delta, x, y);
                if (hl2.distance(o->x, o->y) < hd) {
                    break;
                }
                pd = hd;
                d += FLAGS_adjust_delta;
            }

            float d2 = hl.dir;
            pd = hd;
            for (int i = 0; i < FLAGS_adjust_times; i++) {
                HalfLine hl2(d2 - FLAGS_adjust_delta, x, y);
                if (hl2.distance(o->x, o->y) < hd) {
                    break;
                }
                pd = hd;
                d2 -= FLAGS_adjust_delta;
            }

            bool adjust = false;
            if (d != hl.dir) {
                if (d2 == hl.dir) {
                    adjust = true;
                }
                else {
                    LOG("double Adjust attempt!\n");
                }
            }
            else if (d2 != hl.dir) {
                d = d2;
                adjust = true;
            }

            if (adjust) {
                float gd = distance(gx, gy, x, y);
                float ngx = x + cos(d) * gd;
                float ngy = y + sin(d) * gd;
                LOG("Adjust!: (%.3f %.3f) %.1f => (%.3f %.3f) %.1f\n",
                    gx, gy, rtod(hl.dir), ngx, ngy, rtod(d));
                gx = ngx;
                gy = ngy;
            }

            break;
        }
    }
}

void Rover::proc() {
    if (should_brake) {
        send("b;", true);
    }
    else if (firstTel) {
        guessPos();

        sortObject();

        if (!checkNearGoalMissing()) {
#if 0
            //if (world->updated()) {
            calcBestGoal(&gx, &gy);
            if (gx == 0 && gy == 0) {
#endif
                gx = world->gx;
                gy = world->gy;
#if 0
            }
#endif

            adjustNearOverlap();

            gotoXY(gx, gy);
        }
    }
    else {
        send("a;", true);
    }
}

bool Rover::outOfBounds(float x, float y) const {
    if (x < -dx/2 || x > dx/2 || y < -dy/2 || y > dy/2) {
        return true;
    }
    return false;
}

