#include <map>

#include <google/gflags.h>

#include "world.h"
#include "log.h"
#include "rover.h"
#include "util.h"
#include "geom.h"

DEFINE_double(cannot_pass, 2, "");
DEFINE_double(cannot_pass_c, 3, "");

DEFINE_double(ignore_near_goal_ratio, 1, "");

DEFINE_double(ignore_small_cluster, 1, "");

DEFINE_double(prefer_curdir_ratio, 0.4, "");

DEFINE_double(sharp_turn_braking_speed, 15, "");

static Rover* g_rover;
static World* g_world;

static void showOneCluster(const vector<WO*>& c) {
    for (int j = 0; j < c.size(); j++) {
        const WO* wo = c[j];
        LOG("%d(%c): (%.3f, %.3f) %.3f\n",
            j, wo->type, wo->x, wo->y, wo->r);
    }
}

static float cannot_pass(char t) {
    if (t == 'c') return FLAGS_cannot_pass_c;
    return FLAGS_cannot_pass;
}

static void calcNextPath(float gx, float gy, int cluster_id, int depth,
                         float* gx2, float* gy2);

static bool avoidCluster(int cluster_id, const vector<WO*>& c,
                         float x, float y, float curDir, float minDistance,
                         int depth, float* gx, float* gy) {
    float goalDir = atan2(-y, -x);

    float maxDir = INT_MIN;
    float minDir = INT_MAX;
    float maxAbsDir = INT_MIN;
    float minAbsDir = INT_MAX;
    const WO* maxWo = 0;
    const WO* minWo = 0;

    for (int i = 0; i < c.size(); i++) {
        const WO* wo = c[i];
        float dir = atan2(wo->y - y, wo->x - x);
        //float dd = diffDir(dir, curDir);
        float dd = diffDir(dir, goalDir);

        float d = distance(wo->x, wo->y, x, y);
        float r = wo->r + cannot_pass(wo->type) / 2;
        if (d < r) {
            r -= cannot_pass(wo->type) / 2;
            if (d < r) {
                LOG("ERR: %f vs %f\n", d, r);
                d = r + 0.0000001;
            }
        }

        float t = asin(r / d);

        if (maxDir < dd + t) {
            maxDir = dd + t;
            maxAbsDir = dir + t;
            maxWo = wo;
        }
        if (minDir > dd - t) {
            minDir = dd - t;
            minAbsDir = dir - t;
            minWo = wo;
        }
    }

    const WO* bestWo = 0;
    float bestAbsDir = 0;

    {
        float minR = minWo->r + cannot_pass(minWo->type) * 0.5;
        float minD = distance2(minWo->x, minWo->y, x, y) - minR*minR;
        if (minD >= 0) {
            minD = sqrt(minD);
        }
        else {
            minD = 0;
        }
        float minX = x + cos(minAbsDir) * minD;
        float minY = y + sin(minAbsDir) * minD;

        float maxR = maxWo->r + cannot_pass(maxWo->type) * 0.5;
        float maxD = distance2(maxWo->x, maxWo->y, x, y) - maxR*maxR;
        if (maxD >= 0) {
            maxD = sqrt(maxD);
        }
        else {
            maxD = 0;
        }
        float maxX = x + cos(maxAbsDir) * maxD;
        float maxY = y + sin(maxAbsDir) * maxD;

        if (depth == 0 &&
            maxD + minD > distance(maxX, maxY, minX, minY) * 6) {
            LOG("Long term plan\n");

            float maxX2 = 0;
            float maxY2 = 0;
            calcNextPath(maxX, maxY, cluster_id, depth+1, &maxX2, &maxY2);
            float minX2 = 0;
            float minY2 = 0;
            calcNextPath(minX, minY, cluster_id, depth+1, &minX2, &minY2);
            float maxL2 = length(maxX2, maxY2);
            float minL2 = length(minX2, minY2);
            if (maxL2 < minL2) {
                bestWo = maxWo;
                bestAbsDir = maxAbsDir;
            }
            else if (maxL2 > minL2) {
                bestWo = minWo;
                bestAbsDir = minAbsDir;
            }
        }

        if (g_rover->outOfBounds(minX, minY)) {
            bestWo = maxWo;
            bestAbsDir = maxAbsDir;
        }

        if (g_rover->outOfBounds(maxX, maxY)) {
            bestWo = minWo;
            bestAbsDir = minAbsDir;
        }
    }

    float minGoalRel = absDiffDir(minAbsDir, goalDir);
    float maxGoalRel = absDiffDir(maxAbsDir, goalDir);
    float minCurRel = absDiffDir(minAbsDir, curDir);
    float maxCurRel = absDiffDir(maxAbsDir, curDir);
    LOG("min goal: %.1f max goal: %.1f min cur: %.1f max cur: %.1f\n",
        rtod(minGoalRel), rtod(maxGoalRel), rtod(minCurRel), rtod(maxCurRel));
    if (!bestWo) {
        if (minCurRel + M_PI * FLAGS_prefer_curdir_ratio < maxCurRel) {
            bestWo = minWo;
            bestAbsDir = minAbsDir;
        }
        else if (maxCurRel + M_PI * FLAGS_prefer_curdir_ratio< minCurRel) {
            bestWo = maxWo;
            bestAbsDir = maxAbsDir;
        }
        else if (minGoalRel + minCurRel * 0.1 < maxGoalRel + maxCurRel * 0.1) {
            bestWo = minWo;
            bestAbsDir = minAbsDir;
        }
        else {
            bestWo = maxWo;
            bestAbsDir = maxAbsDir;
        }
    }

    if (distance(bestWo->x, bestWo->y, x, y) - bestWo->r < 1) {
    //if (distance(bestWo->x, bestWo->y, x, y) - bestWo->r < 3) {
        LOG("Too near\n");
        *gx = 0;
        *gy = 0;
        return false;
    }

    float r = bestWo->r + cannot_pass(bestWo->type) * 0.5;
    float d = distance2(bestWo->x, bestWo->y, x, y) - r*r;
    if (d >= 0) {
        d = sqrt(d);
    }
    else {
        d = 0;
    }
    g_world->goal_reason = bestWo;
    *gx = x + cos(bestAbsDir) * d;
    *gy = y + sin(bestAbsDir) * d;

    if (depth < 1) {
        float gx2 = 0;
        float gy2 = 0;
        calcNextPath(*gx, *gy, cluster_id, depth + 1, &gx2, &gy2);
        g_world->gx2 = gx2;
        g_world->gy2 = gy2;
    }

    return true;
}

static void calcNextPath(float gx, float gy, int cluster_id, int depth,
                         float* gx2, float* gy2) {
    HalfLine hl(0, 0, gx, gy);
    int minCluster;
    float minDistance = INT_MAX / 2;
    g_world->chooseCluster(gx, gy, hl, &minCluster, &minDistance,
                           cluster_id);
    *gx2 = 0;
    *gy2 = 0;
    if (minCluster >= 0) {
        const vector<WO*>& c = g_world->cluster[minCluster];
        avoidCluster(minCluster, c, gx, gy, hl.dir /* goalDir */,
                     depth+1, minDistance, gx2, gy2);
    }
}

World::World(Rover* rover0)
    : rover(rover0) {
    g_rover = rover;
    g_world = this;
    reset();
}

void World::reset() {
    added = false;
    gx = 0;
    gy = 0;
    gx2 = 0;
    gy2 = 0;
    prev_cluster = -1;
    goal_reason = 0;
}

void World::addObject(const Obj* o) {
    if (checker.insert(make_pair(make_pair(make_pair(o->x, o->y),
                                           o->r),
                                           o->type)).second) {
        added = true;
        WO* wo = new WO;
        wo->x = o->x;
        wo->y = o->y;
        wo->r = o->r;
        wo->type = o->type;
        wo->clusterId = -1;
        CHECKL(wo->type == 'c' || wo->type == 'b', "%c\n", wo->type);
        wo->id = wos.size();
        wos.push_back(wo);

        int mi = cluster.size();
        for (int i = 0; i < mi; i++) {
            int mj = cluster[i].size();
            for (int j = 0; j < mj; j++) {
                WO* two = cluster[i][j];
                if (distance(wo->x, wo->y, two->x, two->y) - wo->r - two->r <
                    cannot_pass(wo->type)) {
                    if (wo->clusterId == -1) {
                        cluster[i].push_back(wo);
                        wo->clusterId = i;
                    }
                    else if (wo->clusterId != i) {
                        copy(cluster[i].begin(), cluster[i].end(),
                             back_inserter(cluster[wo->clusterId]));
                        cluster.erase(cluster.begin() + i);
                        mi = cluster.size();
                        i--;
                        break;
                    }
                }
            }
        }

        if (wo->clusterId == -1) {
            cluster.push_back(vector<WO*>());
            cluster.back().push_back(wo);
        }
    }
}

void World::chooseCluster(float x, float y, const HalfLine& hl,
                          int* minClusterOut,
                          float* minDistance,
                          int ignore_cluster) {
    int minCluster = -1;

    float goalDistance = length(x, y) - 3;

    for (int i = 0; i < cluster.size(); i++) {
        if (ignore_cluster == i) continue;

        const vector<WO*>& c = cluster[i];
        /*
        if (c.size() == 1 && c[0]->r < FLAGS_ignore_small_cluster) {
            continue;
        }
        */

        for (int j = 0; j < c.size(); j++) {
            const WO* wo = c[j];

            if (fabs(diffDir(atan2(wo->y - y, wo->x - x), hl.dir)) >
                M_PI * 0.5) {
                LOG2("Ignore %d (dir)\n", wo->id);
                LOG2("dirs %.3f %.3f %.3f\n",
                     rtod(atan2(wo->x - x, wo->y - y)), rtod(hl.dir),
                     diffDir(atan2(wo->x - x, wo->y - y), hl.dir));
                continue;
            }

            float d = distance(x, y, wo->x, wo->y);
            if (d > goalDistance * FLAGS_ignore_near_goal_ratio) {
                LOG2("Ignore %d (dis)\n", wo->id);
                continue;
            }

            float ld = hl.distance(wo->x, wo->y) - wo->r;
            if (ld > cannot_pass(wo->type) * 0.5f) {
                LOG2("Ignore %d (line)\n", wo->id);
                continue;
            }

            if (d < *minDistance) {
                *minDistance = d;
                minCluster = i;
            }
        }
    }

    LOG("Min Cluster is %d:\n", minCluster);

    *minClusterOut = minCluster;
}

void World::calcPath(float x, float y, float curDir) {
    HalfLine hl(0, 0, x, y);
    if (fabs(diffDir(hl.dir, curDir)) < 0.2) {
        LOG("Use current dir: %.1f => %.1f\n",
            rtod(hl.dir), rtod(curDir));
        hl = HalfLine(curDir, x, y);
    }

    int minCluster;
    float minDistance = INT_MAX / 2;
    chooseCluster(x, y, hl, &minCluster, &minDistance);

    if (!updated() && distance2(x, y, gx, gy) > 100 &&
        minCluster == prev_cluster) {
        return;
    }
    prev_cluster = minCluster;

    added = false;

    if (minCluster == -1) {
        gx = gy = 0;
        return;
    }

    if (avoidCluster(minCluster, cluster[minCluster],
                     x, y, curDir, minDistance,
                     0, &gx, &gy)) {
#if 0
        if (gx != gx2 || gy != gy2) {
            float dd = absDiffDir(atan2(gy - y, gx - x),
                                  atan2(gy2 - gy, gx2 - gx));
            if (distance(gx, gy, x, y) > 50 &&
                dd > M_PI * 0.5) {
                gx2 = gx;
                gy2 = gy;
                
            }
        }
#endif
    }
    else {
        prev_cluster = -1;
    }
    LOG("Goal (%.3f, %.3f)\n", gx, gy);
}

void World::showClusters() {
    for (int i = 0; i < cluster.size(); i++) {
        LOG("Cluster %d:\n", i);
        showOneCluster(cluster[i]);
    }
}

bool World::needBrake(float x, float y, float speed) const {
    float sharp_turn_braking_speed = FLAGS_sharp_turn_braking_speed;

    if (gx == gx2 && gy == gy2) {
        return false;
    }

    float dd = absDiffDir(atan2(gy - y, gx - x),
                          atan2(gy2 - gy, gx2 - gx));

    float start_brake_distance = 15;
    if (dd > M_PI / 2) {
        start_brake_distance = 30;
        //sharp_turn_braking_speed /= 2;
        LOG("too sharp turn: %f\n", dd);
    }
    start_brake_distance = 15 + 30 * dd / M_PI;
    sharp_turn_braking_speed -= 8 * dd / M_PI;

    if (speed < sharp_turn_braking_speed) {
        return false;
    }

    if (distance(gx, gy, x, y) > start_brake_distance) {
        return false;
    }

    if (goal_reason && goal_reason->r < 1 &&
        prev_cluster >= 0 && cluster[prev_cluster].size() == 1) {
        return false;
    }

    if (dd > dtor(5)) {
        LOG("Need brake: %.1f > %.1f\n", rtod(dd), 5.0f);
        return true;
    }
    return false;
}
